/*
 *                            *
 *                           * *
 *                          *   *
 *                     ***************
 *                      * *       * *
 *                       *  MUMPS  *
 *                      * *       * *
 *                     ***************
 *                          *   *
 *                           * *
 *                            *
 *
 *   shmmgr.h
 *    shared memory manager data structures
 *
 *  
 *   Author: John P. Willis <jpw@coherent-logic.com>
 *    Copyright (C) 1998 MUG Deutschland
 *    Copyright (C) 2021 Coherent Logic Development LLC
 *
 *
 *   This file is part of FreeM.
 *
 *   FreeM is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Affero Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   FreeM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Affero Public License for more details.
 *
 *   You should have received a copy of the GNU Affero Public License
 *   along with FreeM.  If not, see <https://www.gnu.org/licenses/>.
 *
 **/

#if !defined(__shmmgr_h)
# define __shmmgr_h

#include <stdlib.h>
#include <stdint.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/types.h>

/* allocation states */
#define PG_FREE     0
#define PG_ALLOC    1

/* size of a page */
#define PG_SIZE 1024

/* shm_init return values */
#define SHMS_SUCCESS 1
#define SHMS_GET_ERR 0
#define SHMS_ATTACH_ERR -1

/* count of entries in shm directory */
#define SHM_DIRL 10

/* shmd_key types */
#define SHMDK_NONE 0    /* this entry is not in use */
#define SHMDK_JOB 1     /* job table */
#define SHMDK_LCK 2     /* lock table */
#define SHMDK_ZALLO 3   /* zallocate table */ 

/* data structures */

/* shm directory */
typedef struct shm_dir_t {

    /* what does this directory entry represent? */
    short shmd_key;

    /* where does it start? */    
    void *addr;

} shm_dir_t;

/* segment header */
typedef struct shm_hdr_t {

    /* will be set to shm_key to identify whether or not the segment is initialized */
    key_t magic;

    /* pid of the first process that attached to the segment */
    pid_t first_process;

    /* shm address */
    void *shmad;

    /* shm directory */
    shm_dir_t shmdir[SHM_DIRL];
    
} shm_hdr_t;

/* alloc map entry */
typedef struct shm_page_t {

    /* is this the first entry for this allocation? */
    uint8_t  is_first;

    /* can be PG_FREE or PG_ALLOC */
    uint8_t pg_state;

    /* to which pid does this page belong? */
    pid_t pid;

    /* is this the last entry for this allocation? */
    uint8_t is_last;
    
} shm_page_t;

typedef struct shm_config_t {

    /* segment size */
    size_t segsiz;

    /* page size */
    size_t pgsiz;

    /* page count */
    int pgct;

    /* shared memory key */
    key_t key;

    /* segment ID */
    int seg_id;

    /* segment data area */
    void *dta;

    /* pointer to the header */
    shm_hdr_t *hdr;
    
    /* beginning of allocation map */
    shm_page_t *alloc_map;

    /* actual buffer */
    void *buf;
    
} shm_config_t;

/* global variables */
extern shm_config_t *shm_config;

/* function prototypes */
extern short shm_init(const size_t seg_size);
extern shm_page_t *shm_get_alloc_map_entry(const int page_number);
extern void *shm_page_num_to_address(const int page_num);
extern int shm_address_to_page_num(const void *address);
extern void *shm_alloc_pages(const int page_count);
extern void *shm_alloc(const size_t bytes);
extern void shm_free_page(const int page_number);
extern void shm_free(const void *addr);
extern void *shmd_find(const short shmd_key);
extern void *shmd_insert(const short key, const void *addr);
extern void shm_dump(void);
extern void shm_dump_pages(void);
#endif

   
